==> crow-2.15.0.tar.bz2: crow-2.15.0/crow/polytree.h <==
/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2009 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#ifndef CROW_POLYTREE_H
#define CROW_POLYTREE_H

#include "util.h"

namespace Crow {

class PolycellProxy : public Glib::Object {

public:
	~PolycellProxy();

	static PPolycellProxy create(PPolycell polycell);
	PPolycell getPolycell();

	void clear();

private:
	PPolycell polycell;

};

class Polycell : public Gtk::CellEditable {

friend class PolycellRenderer;

public:
	Polycell();
	~Polycell();

	void init(PPolyelem elem, int column);
	void done();

	virtual bool isEditable() { return allowEditing; }

	bool getAllowEditing() { return allowEditing; }
	void setAllowEditing(bool b) { allowEditing= b; }

	bool isFuzzy() { return fuzzy; }
	void setFuzzy(bool fuzzy=true);

	void setAutoStop(bool as) { autoStop= as; }
	bool isEditing() { return editing; }

	PPolyelem getElement();
	Gtk::TreeModel::Path getPath();
	Polytree * getTreeView();
	int getModelColumn() { return column; }

	void startEditing();
	void accept(bool stop=false);
	void reject(bool stop=false);

	typedef sigc::signal<void> SignalAccepted;
	sigc::connection connect_accepted(SignalAccepted::slot_type slot) { return signalAccepted.connect(slot); }

protected:
	void start_editing_vfunc(GdkEvent *);
	void on_remove_widget();

	void stopEditing();
	void editingStarted();
	void editingStopped();
	void accepted() { signalAccepted.emit(); }

	virtual void onEditingStarted() { }
	virtual void onEditingStopped() { }
	virtual bool onAccept() { return true; }
	virtual void onReject() { }

	PolycellWidget * asWidget();

	void invalidate();

	virtual void cellRender(
		Gtk::Widget * widget,
		PGdkWindow window,
		const Gdk::Rectangle & rect,
		Gtk::CellRendererState flags
	) = 0;

	virtual void fuzzyRender(
		Gtk::Widget * widget,
		PGdkWindow window,
		const Gdk::Rectangle & rect,
		Gtk::CellRendererState flags
	) = 0;

private:
	bool autoStop;
	bool editing;
	bool allowEditing;
	PPolyelem element;
	int column;
	bool manualStop;
	SignalAccepted signalAccepted;
	bool fuzzy;

};

class PolycellRenderer : public Gtk::CellRenderer {

public:
	PolycellRenderer(Polytree * tree, int mc);

	Glib::PropertyProxy<PPolycellProxy> property_polycell() { return propertyPolycell.get_proxy(); }

	int getModelColumn() { return column; }

	bool getShowIndents() { return showIndents; }
	void setShowIndents(bool b) { showIndents= b; }

	bool pointInIndent(int depth, const Rectangle & back, int x, int y);
	bool pointInSign(int depth, const Rectangle & back, int x, int y);

protected:
	void get_size_vfunc(
		Gtk::Widget & widget,
		const Gdk::Rectangle * cell_area,
		int * x_offset, int * y_offset,
		int * width, int * height
	) const;

	void render_vfunc(
		const PGdkDrawable & window,
		Gtk::Widget & widget,
		const Gdk::Rectangle & background_area,
		const Gdk::Rectangle & cell_area,
		const Gdk::Rectangle & expose_area,
		Gtk::CellRendererState flags
	);

	Gtk::CellEditable * start_editing_vfunc(
		GdkEvent * event,
		Gtk::Widget & widget,
		const Glib::ustring & path,
		const Gdk::Rectangle & background_area,
		const Gdk::Rectangle & cell_area,
		Gtk::CellRendererState flags
	);

	Polycell * getCurrent();
	bool findNextRowPath(Gtk::TreePath * path);
	void drawSign(
		PGdkWindow window,
		PGdkGC gc,
		bool expanded,
		int depth,
		const Rectangle & clip
	);
	Rectangle getSignRect(const Rectangle & clip, int depth);

	Gdk::Color getBlackColor();
	Gdk::Color getWhiteColor();
	Gdk::Color getGreyColor();

private:
	Glib::Property<PPolycellProxy> propertyPolycell;
	Polytree * treeView;
	int column;
	bool showIndents;

	int signIndent;
	int depthIndent;
	int signSize;

};

struct Polystate : public Object {
	std::list<IdPath> expanded;
	IdPath scroll;
	std::vector<IdPath> selected;
};

typedef Glib::RefPtr<Polystate> PPolystate;

class Polyelem : public Object {

public:
	enum Operation { opRemove, opUpdate, opInsert, opReplace };

	static PPolyelem create(const IdPath & path, PObject object);

	const IdPath * getPath() { return &path; }
	const IdPath * getParent() { return &parent; }

	PObject getObject();
	void setObject(PObject obj);

	Operation getOperation() { return operation; }
	void setOperation(Operation op) { operation= op; }

	const Gtk::TreePath * getTreePath() { return &treePath; }
	void setTreePath(const Gtk::TreePath & treePath) { this->treePath= treePath; }

	Polytree * getTreeView() { return treeView; }
	void setTreeView(Polytree * treeView) { this->treeView= treeView; }

	Gtk::TreeRow getRow();

	PPolycell getValue(int column);
	template <class T> Glib::RefPtr<T> getValue(int column) { return Glib::RefPtr<T>::cast_dynamic(getValue(column)); }
	void setValue(int column, PPolycell polycell);
	void clearValue(int column);

	void clear();

protected:
	Polyelem(const IdPath & path, PObject object);

private:
	IdPath path;
	PObject object;

	IdPath parent;
	Operation operation;

	Polytree * treeView;
	Gtk::TreePath treePath;

};

class Polytree : public Gtk::TreeView {

public:
	typedef std::vector<PObject> Selection;
	typedef sigc::slot<void, PPolyelem> SetRowValues;

	Polytree();

	bool isEditable() { return editable; }
	void setEditable(bool e) { editable= e; }

	bool getImmediateEditing() { return immediateEditing; }
	void setImmediateEditing(bool ie) { immediateEditing= ie; }

	PGtkTreeStore getModel();
	void createModel(int ncol);
	void addColumn(int mcol, const Glib::ustring & name);

	int getColumnWidth(int ncol);
	void setColumnWidth(int ncol, int width);

	Gtk::TreeViewColumn * findViewColumn(int modelColumn);

	void setSetRowValues(SetRowValues setRowValues) { this->setRowValues= setRowValues; }

	void prepare();
	void update(const IdPath & path, PObject object);
	void submit();

	const Selection * getSelection() { return &selection; }
	void setSelection(const Selection & selection, bool fire);

	std::vector<Gtk::TreePath> getTreePathSelection();
	void setTreePathSelection(const std::vector<Gtk::TreePath> & selection, bool fire);

	PObject saveState();
	void loadState(PObject state);

	void expandRow(PObject obj);
	void expandAll();
	PObject getByTreePath(const Gtk::TreePath & path);

	bool isEditing() { return editingPolycell != NULL; }
	void startEditing(const Gtk::TreePath & path, int modelColumn);
	Polycell * getEditingPolycell() { return editingPolycell; }

	void editingStarted(Polycell * polycell);
	void editingStopped(Polycell * polycell);
	void invalidate(const Gtk::TreePath & path, int column);

	void acceptEditing();
	void rejectEditing();

	typedef sigc::signal<void> SignalContextMenu;
	typedef sigc::signal<void> SignalSelectionChanged;
	typedef sigc::signal<void> SignalSubmited;

	sigc::connection connect_context_menu(SignalContextMenu::slot_type slot) { return signalContextMenu.connect(slot); }
	sigc::connection connect_selection_changed(SignalSelectionChanged::slot_type slot) { return signalSelectionChanged.connect(slot); }
	sigc::connection connect_submited(SignalSubmited::slot_type slot) { return signalSubmited.connect(slot); }

protected:
	bool on_button_press_event(GdkEventButton * event);
	bool on_button_release_event(GdkEventButton * event);
	void on_size_allocate(Gtk::Allocation & allocation);
	void on_row_expanded(const Gtk::TreeModel::iterator & iter, const Gtk::TreeModel::Path & path);
	void on_row_collapsed(const Gtk::TreeModel::iterator & iter, const Gtk::TreeModel::Path & path);

	void onSelectionChanged();

	typedef std::map<IdPath, PPolyelem> Elements;

	Elements::iterator process(
		const Gtk::TreeNodeChildren & tnc,
		Elements::iterator el,
		const IdPath & parent
	);
	void initElem(PPolyelem E, const Gtk::TreeIter & rowit);

	Gtk::TreePath getEditingPath();
	void select(const Selection & selection, bool fire, bool force=false);
	void selectPathes(const std::vector<Gtk::TreePath> & pathes);
	void expandToPath(const Gtk::TreePath & path);
	bool rowVisible(const Gtk::TreePath & path);
	void showSelection();

	PPolyelem findElement(const Gtk::TreePath & treePath);
	PPolyelem findElement(PObject object);
	PPolyelem findElement(const IdPath & path);

	void dump(const std::string & str);

private:
	SignalContextMenu signalContextMenu;
	SignalSelectionChanged signalSelectionChanged;
	SignalSubmited signalSubmited;

	bool editable;
	bool immediateEditing;
	bool lock;
	Selection selection;
	Polycell * editingPolycell;

	std::vector<IdPath> keepSelected;
	int keepEditing;

	Elements elements;
	SetRowValues setRowValues;

};

class PolycellText : public virtual Gtk::Alignment, public virtual Polycell {

public:
	PolycellText();

	const Glib::ustring * getText() { return &text; }
	virtual void setText(const Glib::ustring & text);

	void setEllipsize(Pango::EllipsizeMode emode) { ellipsize= emode; }

	template <class T> void changeAttribute(const T & attribute) { ChangeAttribute(&attributes, attribute); }

protected:
	PPangoLayout createLayout(const Glib::ustring & str, Gtk::Widget * widget, int width=-1);
	PPangoLayout getLayout(Gtk::Widget * widget, int width=-1);

	void textRender(
		PPangoLayout layout,
		Gtk::Widget * widget,
		PGdkWindow window,
		const Gdk::Rectangle & rect,
		Gtk::CellRendererState flags
	);

	void cellRender(
		Gtk::Widget * widget,
		PGdkWindow window,
		const Gdk::Rectangle & rect,
		Gtk::CellRendererState flags
	);

	void fuzzyRender(
		Gtk::Widget * widget,
		PGdkWindow window,
		const Gdk::Rectangle & rect,
		Gtk::CellRendererState flags
	);

private:
	Glib::ustring text;
	Pango::AttrList attributes;
	Pango::EllipsizeMode ellipsize;

};

class PolycellWidget : public PolycellText {

public:
	PolycellWidget();

	virtual void setPlacement(const Rectangle & rect, int hshift, int vshift);

protected:
	void on_size_allocate(Gtk::Allocation & allocation);

private:
	Rectangle placement;
	int hShift, vShift;

};

class PolycellInput : public PolycellWidget {

public:
	typedef sigc::slot<bool, Glib::ustring, Glib::ustring> Validator;

	PolycellInput();

	void setText(const Glib::ustring & text);
	void setEntryEditable(bool ee);
	void setValidator(const Validator & val) { validator= val; }
	Glib::ustring getEntryText();

	bool isPopupEnabled();
	void setPopupEnabled(bool enabled);

	void setPlacement(const Rectangle & rect, int hshift, int vshift);

	typedef sigc::signal<void> SignalPopup;
	sigc::connection connect_popup(SignalPopup::slot_type slot) { return signalPopup.connect(slot); }

protected:
	void onEditingStarted();
	bool onAccept();
	void onReject();

	void onEntryActivate();
	bool onEntryKeyPressEvent(GdkEventKey * event);
	void onButtonClicked() { signalPopup.emit(); }

	bool validatorDefault(const Glib::ustring & oldtext, const Glib::ustring & newtext) { return true; }

private:
	Gtk::Button button;
	Gtk::Entry entry;
	Gtk::HBox hBox;

	SignalPopup signalPopup;
	Validator validator;

};

class PolycellChoice : public PolycellWidget {

public:
	PolycellChoice();

	void setText(const Glib::ustring & text);

	int getChoice() { return choice; }
	void setChoice(int choice);

	void addChoice(const Glib::ustring & text);

protected:
	void onEditingStarted();
	bool onAccept();

	void onComboChanged();

private:
	Gtk::ComboBoxText combo;

	bool lock;
	int choice;
	std::vector<Glib::ustring> choices;

};

}

#endif

==> crow-2.17.1.tar.bz2: crow-2.17.1/crow/polytree.h <==
/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#ifndef CROW_POLYTREE_H
#define CROW_POLYTREE_H

#include "util.h"

namespace Crow {

class PolycellProxy : public Glib::Object {

public:
	~PolycellProxy();

	static PPolycellProxy create(PPolycell polycell);
	PPolycell getPolycell();

	void clear();

private:
	PPolycell polycell;

};

class Polycell : public Gtk::CellEditable {

friend class PolycellRenderer;

public:
	Polycell();
	~Polycell();

	void init(PPolyelem elem, int column);
	void done();

	virtual bool isEditable() { return allowEditing; }

	bool getAllowEditing() { return allowEditing; }
	void setAllowEditing(bool b) { allowEditing= b; }

	bool isFuzzy() { return fuzzy; }
	void setFuzzy(bool fuzzy=true);

	void setAutoStop(bool as) { autoStop= as; }
	bool isEditing() { return editing; }

	PPolyelem getElement();
	Gtk::TreeModel::Path getPath();
	Polytree * getTreeView();
	int getModelColumn() { return column; }

	void startEditing();
	void accept(bool stop=false);
	void reject(bool stop=false);

	typedef sigc::signal<void> SignalAccepted;
	sigc::connection connect_accepted(SignalAccepted::slot_type slot) { return signalAccepted.connect(slot); }

protected:
	void start_editing_vfunc(GdkEvent *);
	void on_remove_widget();

	void stopEditing();
	void editingStarted();
	void editingStopped();
	void accepted() { signalAccepted.emit(); }

	virtual void onEditingStarted() { }
	virtual void onEditingStopped() { }
	virtual bool onAccept() { return true; }
	virtual void onReject() { }

	PolycellWidget * asWidget();

	void invalidate();

	virtual void cellRender(
		Gtk::Widget * widget,
		PGdkWindow window,
		const Gdk::Rectangle & rect,
		Gtk::CellRendererState flags
	) = 0;

	virtual void fuzzyRender(
		Gtk::Widget * widget,
		PGdkWindow window,
		const Gdk::Rectangle & rect,
		Gtk::CellRendererState flags
	) = 0;

private:
	bool autoStop;
	bool editing;
	bool allowEditing;
	PPolyelem element;
	int column;
	bool manualStop;
	SignalAccepted signalAccepted;
	bool fuzzy;

};

class PolycellRenderer : public Gtk::CellRenderer {

public:
	PolycellRenderer(Polytree * tree, int mc);

	Glib::PropertyProxy<PPolycellProxy> property_polycell() { return propertyPolycell.get_proxy(); }

	int getModelColumn() { return column; }

	bool getShowIndents() { return showIndents; }
	void setShowIndents(bool b) { showIndents= b; }

	bool pointInIndent(int depth, const Rectangle & back, int x, int y);
	bool pointInSign(int depth, const Rectangle & back, int x, int y);

protected:
	void get_size_vfunc(
		Gtk::Widget & widget,
		const Gdk::Rectangle * cell_area,
		int * x_offset, int * y_offset,
		int * width, int * height
	) const;

	void render_vfunc(
		const PGdkDrawable & window,
		Gtk::Widget & widget,
		const Gdk::Rectangle & background_area,
		const Gdk::Rectangle & cell_area,
		const Gdk::Rectangle & expose_area,
		Gtk::CellRendererState flags
	);

	Gtk::CellEditable * start_editing_vfunc(
		GdkEvent * event,
		Gtk::Widget & widget,
		ustring_cr path,
		const Gdk::Rectangle & background_area,
		const Gdk::Rectangle & cell_area,
		Gtk::CellRendererState flags
	);

	Polycell * getCurrent();
	bool findNextRowPath(Gtk::TreePath * path);
	void drawSign(
		PGdkWindow window,
		PGdkGC gc,
		bool expanded,
		int depth,
		const Rectangle & clip
	);
	Rectangle getSignRect(const Rectangle & clip, int depth);

	Gdk::Color getBlackColor();
	Gdk::Color getWhiteColor();
	Gdk::Color getGreyColor();

private:
	Glib::Property<PPolycellProxy> propertyPolycell;
	Polytree * treeView;
	int column;
	bool showIndents;

	int signIndent;
	int depthIndent;
	int signSize;

};

struct Polystate : public Object {
	std::list<IdPath> expanded;
	IdPath scroll;
	std::vector<IdPath> selected;
};

typedef Glib::RefPtr<Polystate> PPolystate;

class Polyelem : public Object {

public:
	enum Operation { opRemove, opUpdate, opInsert, opReplace };

	static PPolyelem create(const IdPath & path, PObject object);

	const IdPath * getPath() { return &path; }
	const IdPath * getParent() { return &parent; }

	PObject getObject();
	void setObject(PObject obj);

	Operation getOperation() { return operation; }
	void setOperation(Operation op) { operation= op; }

	const Gtk::TreePath * getTreePath() { return &treePath; }
	void setTreePath(const Gtk::TreePath & treePath) { this->treePath= treePath; }

	Polytree * getTreeView() { return treeView; }
	void setTreeView(Polytree * treeView) { this->treeView= treeView; }

	Gtk::TreeRow getRow();

	PPolycell getValue(int column);
	template <class T> Glib::RefPtr<T> getValue(int column) { return Glib::RefPtr<T>::cast_dynamic(getValue(column)); }
	void setValue(int column, PPolycell polycell);
	void clearValue(int column);

	void clear();

protected:
	Polyelem(const IdPath & path, PObject object);

private:
	IdPath path;
	PObject object;

	IdPath parent;
	Operation operation;

	Polytree * treeView;
	Gtk::TreePath treePath;

};

class Polytree : public Gtk::TreeView {

public:
	typedef std::vector<PObject> Selection;
	typedef sigc::slot<void, PPolyelem> SetRowValues;

	Polytree();

	bool isEditable() { return editable; }
	void setEditable(bool e) { editable= e; }

	bool getImmediateEditing() { return immediateEditing; }
	void setImmediateEditing(bool ie) { immediateEditing= ie; }

	PGtkTreeStore getModel();
	void createModel(int ncol);
	void addColumn(int mcol, ustring_cr name);

	int getColumnWidth(int ncol);
	void setColumnWidth(int ncol, int width);

	Gtk::TreeViewColumn * findViewColumn(int modelColumn);

	void setSetRowValues(SetRowValues setRowValues) { this->setRowValues= setRowValues; }

	void prepare();
	void update(const IdPath & path, PObject object);
	void submit();

	const Selection * getSelection() { return &selection; }
	void setSelection(const Selection & selection, bool fire);

	std::vector<Gtk::TreePath> getTreePathSelection();
	void setTreePathSelection(const std::vector<Gtk::TreePath> & selection, bool fire);

	PObject saveState();
	void loadState(PObject state);

	void expandRow(PObject obj);
	void expandAll();
	PObject getByTreePath(const Gtk::TreePath & path);

	bool isEditing() { return editingPolycell != NULL; }
	void startEditing(const Gtk::TreePath & path, int modelColumn);
	Polycell * getEditingPolycell() { return editingPolycell; }

	void editingStarted(Polycell * polycell);
	void editingStopped(Polycell * polycell);
	void invalidate(const Gtk::TreePath & path, int column);

	void acceptEditing();
	void rejectEditing();

	typedef sigc::signal<void> SignalContextMenu;
	typedef sigc::signal<void> SignalSelectionChanged;
	typedef sigc::signal<void> SignalSubmited;

	sigc::connection connect_context_menu(SignalContextMenu::slot_type slot) { return signalContextMenu.connect(slot); }
	sigc::connection connect_selection_changed(SignalSelectionChanged::slot_type slot) { return signalSelectionChanged.connect(slot); }
	sigc::connection connect_submited(SignalSubmited::slot_type slot) { return signalSubmited.connect(slot); }

protected:
	bool on_button_press_event(GdkEventButton * event);
	bool on_button_release_event(GdkEventButton * event);
	void on_size_allocate(Gtk::Allocation & allocation);
	void on_row_expanded(const Gtk::TreeModel::iterator & iter, const Gtk::TreeModel::Path & path);
	void on_row_collapsed(const Gtk::TreeModel::iterator & iter, const Gtk::TreeModel::Path & path);

	void onSelectionChanged();

	typedef std::map<IdPath, PPolyelem> Elements;

	Elements::iterator process(
		const Gtk::TreeNodeChildren & tnc,
		Elements::iterator el,
		const IdPath & parent
	);
	void initElem(PPolyelem E, const Gtk::TreeIter & rowit);

	Gtk::TreePath getEditingPath();
	void select(const Selection & selection, bool fire, bool force=false);
	void selectPathes(const std::vector<Gtk::TreePath> & pathes);
	void expandToPath(const Gtk::TreePath & path);
	bool rowVisible(const Gtk::TreePath & path);
	void showSelection();

	PPolyelem findElement(const Gtk::TreePath & treePath);
	PPolyelem findElement(PObject object);
	PPolyelem findElement(const IdPath & path);

	void dump(string_cr str);

private:
	SignalContextMenu signalContextMenu;
	SignalSelectionChanged signalSelectionChanged;
	SignalSubmited signalSubmited;

	bool editable;
	bool immediateEditing;
	bool lock;
	Selection selection;
	Polycell * editingPolycell;

	std::vector<IdPath> keepSelected;
	int keepEditing;

	Elements elements;
	SetRowValues setRowValues;

};

class CustomAlignment : public Gtk::Alignment {

public:
	CustomAlignment();

private:
	Glib::Property<bool> propertyEditingCanceled;

};

class PolycellText : public virtual CustomAlignment, public virtual Polycell {

public:
	PolycellText();

	const Glib::ustring * getText() { return &text; }
	virtual void setText(ustring_cr text);

	void setEllipsize(Pango::EllipsizeMode emode) { ellipsize= emode; }

	template <class T> void changeAttribute(const T & attribute) { ChangeAttribute(&attributes, attribute); }

protected:
	PPangoLayout createLayout(ustring_cr str, Gtk::Widget * widget, int width=-1);
	PPangoLayout getLayout(Gtk::Widget * widget, int width=-1);

	void textRender(
		PPangoLayout layout,
		Gtk::Widget * widget,
		PGdkWindow window,
		const Gdk::Rectangle & rect,
		Gtk::CellRendererState flags
	);

	void cellRender(
		Gtk::Widget * widget,
		PGdkWindow window,
		const Gdk::Rectangle & rect,
		Gtk::CellRendererState flags
	);

	void fuzzyRender(
		Gtk::Widget * widget,
		PGdkWindow window,
		const Gdk::Rectangle & rect,
		Gtk::CellRendererState flags
	);

private:
	Glib::ustring text;
	Pango::AttrList attributes;
	Pango::EllipsizeMode ellipsize;

};

class PolycellWidget : public PolycellText {

public:
	PolycellWidget();

	virtual void setPlacement(const Rectangle & rect, int hshift, int vshift);

protected:
	void on_size_allocate(Gtk::Allocation & allocation);

private:
	Rectangle placement;
	int hShift, vShift;

};

class PolycellInput : public PolycellWidget {

public:
	typedef sigc::slot<bool, Glib::ustring, Glib::ustring> Validator;

	PolycellInput();

	void setText(ustring_cr text);
	void setEntryEditable(bool ee);
	void setValidator(const Validator & val) { validator= val; }
	Glib::ustring getEntryText();

	bool isPopupEnabled();
	void setPopupEnabled(bool enabled);

	void setPlacement(const Rectangle & rect, int hshift, int vshift);

	typedef sigc::signal<void> SignalPopup;
	sigc::connection connect_popup(SignalPopup::slot_type slot) { return signalPopup.connect(slot); }

protected:
	void onEditingStarted();
	bool onAccept();
	void onReject();

	void onEntryActivate();
	bool onEntryKeyPressEvent(GdkEventKey * event);
	void onButtonClicked() { signalPopup.emit(); }

	bool validatorDefault(ustring_cr oldtext, ustring_cr newtext) { return true; }

private:
	Gtk::Button button;
	Gtk::Entry entry;
	Gtk::HBox hBox;

	SignalPopup signalPopup;
	Validator validator;

};

class PolycellChoice : public PolycellWidget {

public:
	PolycellChoice();

	void setText(ustring_cr text);

	int getChoice() { return choice; }
	void setChoice(int choice);

	void addChoice(ustring_cr text);

protected:
	void onEditingStarted();
	bool onAccept();

	void onComboChanged();

private:
	Gtk::ComboBoxText combo;

	bool lock;
	int choice;
	std::vector<Glib::ustring> choices;

};

}

#endif
